# $Id: latlong2utm.py 112 2005-04-21 18:50:29Z jcea $

from math import sin,cos,tan,sqrt,pi

# WGS-84
EquatorialRadius=6378137
eccSquared=eccentricitySquared=0.00669438

deg2rad=2*pi/360
rad2deg=1/deg2rad

letraUTM="ZZZZZZZZZZCCCCCCCCDDDDDDDDEEEEEEEEFFFFFFFFGGGGGGGG" \
         "HHHHHHHHJJJJJJJJKKKKKKKKLLLLLLLLMMMMMMMMNNNNNNNN" \
         "PPPPPPPPQQQQQQQQRRRRRRRRSSSSSSSSTTTTTTTTUUUUUUUU" \
         "VVVVVVVVWWWWWWWWXXXXXXXXXXXXZZZZZZ"

def latlong2utm(lat,long) :
  lat,lat2=float(lat[:-1]),lat[-1]
  long,long2=float(long[:-1]),long[-1]
  lat=lat//100.0+(lat%100.0)/60.0
  long=long//100.0+(long%100.0)/60.0
  if lat2!="N" : lat=-lat
  if long2!="E" : long=-long

  latRad=lat*deg2rad
  longRad=long*deg2rad

  zona=int((long+180)/6)+1

  if (lat>=56.0) and (lat<64.0) and (long>=3.0) and (long<12.0) :
    zona=32

  if (lat>=72.0) and (lat<84.0) :
    if (long>=0.0) and (long<9.0) :
      zona=32
    elif (long>=9.0) and (long<21.0) :
      zona=33
    elif (long>=21.0) and (long<33.0) :
      zona=35
    elif (long>=33.0) and (long<42.0) :
      zona=37

  longOrigen=(zona-1)*6-180+3
  longOrigenRad=longOrigen*deg2rad

  eccPrimeSquared = (eccSquared)/(1-eccSquared)

  N=EquatorialRadius/sqrt(1-eccSquared*sin(latRad)*sin(latRad))
  T=tan(latRad)*tan(latRad)
  C=eccPrimeSquared*cos(latRad)*cos(latRad)
  A=cos(latRad)*(longRad-longOrigenRad)
  M=EquatorialRadius*((1-eccSquared/4-3*eccSquared*eccSquared/64-5*eccSquared*eccSquared*eccSquared/256)*latRad
    -(3*eccSquared/8+3*eccSquared*eccSquared/32+45*eccSquared*eccSquared*eccSquared/1024)*sin(2*latRad)
    +(15*eccSquared*eccSquared/256+45*eccSquared*eccSquared*eccSquared/1024)*sin(4*latRad)
    -(35*eccSquared*eccSquared*eccSquared/3072)*sin(6*latRad))

  UTMEasting=0.9996*N*(A+(1-T+C)*A*A*A/6 \
             +(5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120) \
             +500000.0

  UTMNorthing=0.9996*(M+N*tan(latRad)*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24 \
              + (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720))

  if lat<0.0 :
    UTMNorthing+=10000000.0

  return zona,letraUTM[int(90+lat)],UTMEasting,UTMNorthing

def utm2latlong(utm) :
  UTMZone,UTMEasting,UTMNorthing=utm.split()
  x=float(UTMEasting)-500000.0
  y=float(UTMNorthing)
  zona,letra=int(UTMZone[:-1]),UTMZone[-1]
  if letra<"N" :
    y-=10000000.0

  e1=(1-sqrt(1-eccSquared))/(1+sqrt(1-eccSquared))

  LongOrigen=(zona-1)*6-180+3
  eccPrimeSquared=(eccSquared)/(1-eccSquared)
  M=y/0.9996
  mu = M/(EquatorialRadius*(1-eccSquared/4-3*eccSquared*eccSquared/64-5*eccSquared*eccSquared*eccSquared/256))
  phi1Rad = mu	+ (3*e1/2-27*e1*e1*e1/32)*sin(2*mu) \
            + (21*e1*e1/16-55*e1*e1*e1*e1/32)*sin(4*mu) \
            +(151*e1*e1*e1/96)*sin(6*mu)
  phi1 = phi1Rad*rad2deg
  N1 = EquatorialRadius/sqrt(1-eccSquared*sin(phi1Rad)*sin(phi1Rad))
  T1 = tan(phi1Rad)*tan(phi1Rad)
  C1 = eccPrimeSquared*cos(phi1Rad)*cos(phi1Rad)
  R1 = EquatorialRadius*(1-eccSquared)/pow(1-eccSquared*sin(phi1Rad)*sin(phi1Rad), 1.5)
  D = x/(N1*0.9996)
  Lat = phi1Rad - (N1*tan(phi1Rad)/R1)*(D*D/2-(5+3*T1+10*C1-4*C1*C1-9*eccPrimeSquared)*D*D*D*D/24 \
        +(61+90*T1+298*C1+45*T1*T1-252*eccPrimeSquared-3*C1*C1)*D*D*D*D*D*D/720)
  Lat = Lat * rad2deg
  Long = (D-(1+2*T1+C1)*D*D*D/6+(5-2*C1+28*T1-3*C1*C1+8*eccPrimeSquared+24*T1*T1) \
         *D*D*D*D*D/120)/cos(phi1Rad)
  Long = LongOrigen + Long * rad2deg

  lat3,long3="N","E"
  if Lat<0.0 :
    lat3="S"
    Lat=-Lat
  if Long<0.0 :
    long3="W"
    Long=-Long

  lat2,long2=int(Lat),int(Long)
  return str(round(lat2*100+(Lat-lat2)*60,4))+lat3, \
         str(round(long2*100+(Long-long2)*60,4))+long3


if __name__=="__main__" :
  import sys

  if sys.argv[0]=="latlong2utm.py" :
    if len(sys.argv) !=3 :
      print "Formato: %s lat[N|S] long[W|E]" %sys.argv[0]
      sys.exit(1)

    lat,long=sys.argv[1],sys.argv[2]
    a,b,c,d=latlong2utm(lat,long)
    print "%02d%s %06d %07.d" %(a,b,int(round(c)),int(round(d)))
    sys.exit(0)

  if sys.argv[0]=="utm2latlong.py" :
    if len(sys.argv) !=4 :
      print "Formato: %s UTMZone UTMEasting UTMNorthing" %sys.argv[0]
      sys.exit(1)

    lat,long=utm2latlong(" ".join(sys.argv[1:4]))
    print lat,long
    sys.exit(0)

  print "Este fichero DEBE llamarse 'latlong2utm.py' o 'utm2latlong.py'"
  sys.exit(1)

